------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--               OCARINA.AADL.PARSER.COMPONENTS.CONNECTIONS                 --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--               Copyright (C) 2004-2007, GET-Telecom Paris.                --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

with Locations;
with Ocarina.Nodes;
with Ocarina.AADL.Lexer;
with Ocarina.AADL.Tokens;
with Ocarina.AADL.Parser.Identifiers;
with Ocarina.AADL.Parser.Components.Modes;
with Ocarina.AADL.Parser.Properties;
with Ocarina.Builder.Components.Connections;
with Ocarina.Entities.Components.Connections;

package body Ocarina.AADL.Parser.Components.Connections is

   ------------------
   -- P_Connection --
   ------------------

   --  connection ::=
   --     port_connection | parameter_connection | access_connection

   --  connection_refinement ::=   port_connection_refinement
   --                            | parameter_connection_refinement
   --                            | access_connection_refinement

   --  port_connection ::=   data_connection | event_connection
   --                      | event_data_connection | port_group_connection

   --  data_connection ::= [ defining_data_connection_identifier :]
   --     data port source_unique_port_identifier
   --               ( immediate_connection_symbol | delayed_connection_symbol )
   --               destination_unique_port_identifier
   --        [ { { property_association }+ } ] [ in_modes_and_transitions ] ;

   --  immediate_connection_symbol ::= ->>
   --  delayed_connection_symbol   ::= ->

   --  event_connection ::= [ defining_event_connection_identifier :]
   --     event port source_unique_port_identifier
   --             -> destination_unique_port_identifier
   --        [ { { property_association }+ } ] [ in_modes_and_transitions ] ;

   --  event_data_connection ::= [ defining_event_data_connection_identifier :]
   --     event data port source_unique_port_identifier
   --                  -> destination_unique_port_identifier
   --        [ { { property_association }+ } ] [ in_modes_and_transitions ] ;

   --  port_group_connection ::= [ defining_port_group_connection_identifier :]
   --     port group source_unique_port_group_identifier
   --             -> destination_unique_port_group_identifier
   --        [ { { property_association }+ } ] [ in_modes_and_transitions ] ;

   --  port_connection_refinement ::=
   --     connection_identifier : refined to
   --     ( data port | event port | event data port | port group )
   --        ( ( { { property_association }+ } [ in_modes_and_transitions ] )
   --          | in_modes_and_transitions ) ;

   --  parameter_connection ::= [ defining_parameter_connection_identifier :]
   --     parameter source_unique_parameter_identifier
   --            -> destination_unique_parameter_identifier
   --        [ { { property_association }+ } ] [ in_modes ] ;

   --  parameter_connection_refinement ::=
   --     connection_identifier : refined to parameter
   --       { { property_association }+ } [ in_modes ] ;

   --  access_connection ::= [ access_connection_identifier :]
   --     ( bus | data ) access unique_access provider_identifier
   --                        -> unique_access_requirer_identifier
   --       [ { { property_association }+ } ] [ in_modes ] ;

   --  access_connection_refinement ::=
   --     connection_identifier : refined to ( bus | data ) access
   --       { { property_association }+ } [ in_modes ] ;

   function P_Connection
     (Container : Types.Node_Id;
      Refinable : Boolean)
     return Types.Node_Id
   is
      use Locations;
      use Ocarina.Nodes;
      use Lexer;
      use Tokens;
      use Ocarina.AADL.Parser.Identifiers;
      use Ocarina.AADL.Parser.Components.Modes;
      use Ocarina.AADL.Parser.Properties;
      use Ocarina.Builder.Components.Connections;
      use Ocarina.Entities.Components.Connections;

      Connection      : Node_Id;
      Identifier      : Node_Id;
      Is_Refinement   : Boolean;
      Category        : Connection_Type;
      Source          : Node_Id := No_Node;
      Destination     : Node_Id := No_Node;
      In_Modes        : Node_Id;
      OK              : Boolean;
      Code            : Parsing_Code;
      Loc             : Location;
      Start_Loc       : Location := No_Location;
      Properties_Loc  : Location;

   begin
      P_Identifier_Refined_To (Refinable_To_RT (Refinable), True,
                               PC_Connection, PC_Connection_Refinement,
                               T_Semicolon, Identifier, Is_Refinement, OK);
      if not OK then
         return No_Node;
      end if;

      if Present (Identifier) then
         Start_Loc := Ocarina.Nodes.Loc (Identifier);   --  update Start_Loc
      end if;

      if Is_Refinement then
         Code := PC_Connection_Refinement;
      else
         Code := PC_Connection;
      end if;

      --  parsing connection category
      Save_Lexer (Loc);
      Scan_Token;

      if Start_Loc = No_Location then
         Start_Loc := Token_Location;
      end if;

      case Token is
         when T_Data =>        --  data_connection or access_connection
            Scan_Token;
            case Token is
               when T_Port =>      --  data_connection
                  Category := CT_Data;
                  if Is_Refinement then
                     Code := PC_Data_Connection_Refinement;
                  else
                     Code := PC_Data_Connection;
                  end if;

               when T_Access =>    --  access_connection
                  Category := CT_Access_Data;
                  if Is_Refinement then
                     Code := PC_Access_Connection_Refinement;
                  else
                     Code := PC_Access_Connection;
                  end if;

               when others =>
                  DPE (Code, (T_Port, T_Access));
                  Skip_Tokens (T_Semicolon);
                  return No_Node;
            end case;

         when T_Event =>       --  event_connection or event_data_connection
            Scan_Token;
            case Token is
               when T_Data =>      --  event_data_connection
                  if Is_Refinement then
                     Code := PC_Event_Data_Connection_Refinement;
                  else
                     Code := PC_Event_Data_Connection;
                  end if;

                  Scan_Token;
                  if Token /= T_Port then
                     DPE (Code, T_Port);
                     Skip_Tokens (T_Semicolon);
                     return No_Node;
                  end if;

                  Category := CT_Event_Data;

               when T_Port =>      --  event_connection
                  Category := CT_Event;
                  if Is_Refinement then
                     Code := PC_Event_Connection_Refinement;
                  else
                     Code := PC_Event_Connection;
                  end if;

               when others =>
                  DPE (Code, (T_Data, T_Port));
                  Skip_Tokens (T_Semicolon);
                  return No_Node;
            end case;

         when T_Port =>        --  port_group_connection
            if Is_Refinement then
               Code := PC_Port_Group_Connection_Refinement;
            else
               Code := PC_Port_Group_Connection;
            end if;

            Scan_Token;
            if Token /= T_Group then
               DPE (Code, T_Group);
               Skip_Tokens (T_Semicolon);
               return No_Node;
            end if;

            Category := CT_Port_Group;

         when T_Parameter =>   --  parameter_connection
            Category := CT_Parameter;
            if Is_Refinement then
               Code := PC_Parameter_Connection_Refinement;
            else
               Code := PC_Parameter_Connection;
            end if;

         when T_Bus =>         --  access_connection
            if Is_Refinement then
               Code := PC_Access_Connection_Refinement;
            else
               Code := PC_Access_Connection;
            end if;

            Scan_Token;
            if Token /= T_Access then
               DPE (Code, T_Access);
               Skip_Tokens (T_Semicolon);
               return No_Node;
            end if;

            Category := CT_Access_Bus;

         when T_Subprogram =>
            --  subprogram access (not in AADL 1.0)

            if Is_Refinement then
               Code := PC_Access_Connection_Refinement;
            else
               Code := PC_Access_Connection;
            end if;

            Scan_Token;
            if Token /= T_Access then
               DPE (Code, T_Access);
               Skip_Tokens (T_Semicolon);
               return No_Node;
            end if;

            Category := CT_Access_Subprogram;

         when others =>
            if Present (Identifier) then
               DPE (Code, (T_Data, T_Event, T_Bus, T_Port, T_Parameter));
               Skip_Tokens (T_Semicolon);
               return No_Node;
            else
               --  nothing was parsed, try to parse other stuff
               Restore_Lexer (Loc);
               return No_Node;
            end if;
      end case;

      --  parse connection source et destination

      if not Is_Refinement then
         Source := P_Entity_Reference (Code);

         if No (Source) then
            Skip_Tokens (T_Semicolon);
            return No_Node;
         end if;

         Scan_Token;
         if Token /= T_Connection then
            if Token = T_Delayed_Connection and then Category = CT_Data then
               Category := CT_Data_Delayed;
            else
               if Category = CT_Data then
                  DPE (Code, (T_Connection, T_Delayed_Connection));
               else
                  DPE (Code, T_Connection);
               end if;
               Skip_Tokens (T_Semicolon);
               return No_Node;
            end if;
         end if;

         Destination := P_Entity_Reference (Code);

         if No (Destination) then
            Skip_Tokens (T_Semicolon);
            return No_Node;
         end if;
      end if;

      --  parse properties and in_modes

      Save_Lexer (Properties_Loc);
      --  We save the location of the properties, as we have to parse
      --  it later (after the connection node has been created)

      Connection := Add_New_Connection (Loc           => Start_Loc,
                                        Name          => Identifier,
                                        Comp_Impl     => Container,
                                        Category      => Category,
                                        Destination   => Destination,
                                        Source        => Source,
                                        Is_Refinement => Is_Refinement,
                                        In_Modes      => No_Node);

      OK := P_Property_Associations (Connection, True, PAT_Simple, Code);

      if not OK then
         Skip_Tokens (T_Semicolon);
         return No_Node;
      end if;

      Save_Lexer (Loc);
      Scan_Token;
      if No (Ocarina.Nodes.Properties (Connection)) then
         --  Property_Associations are not defined,
         --  In_Modes_And_Transitions must exist if Is_Refinement = TRUE

         if Is_Refinement and then Token /= T_In then
            DPE (Code, (T_In, T_Left_Curly_Bracket));
            Skip_Tokens (T_Semicolon);
            return No_Node;
         end if;
      end if;

      if Token = T_In then
         case Category is
            when CT_Data
              | CT_Data_Delayed
              | CT_Event_Data
              | CT_Event
              | CT_Port_Group =>
               In_Modes := P_In_Modes (PC_In_Modes_And_Transitions);

            when CT_Parameter
              | CT_Access_Bus
              | CT_Access_Data
              | CT_Access_Subprogram =>
               In_Modes := P_In_Modes (PC_In_Modes);
         end case;

         if No (In_Modes) then
            --  error when parsing In_Modes, quit

            Skip_Tokens (T_Semicolon);
            return No_Node;
         else
            Set_In_Modes (Connection, In_Modes);
         end if;
      else
         In_Modes := No_Node;
         Restore_Lexer (Loc);
      end if;

      Save_Lexer (Loc);
      Scan_Token;
      if Token /= T_Semicolon then
         DPE (Code, T_Semicolon);
         Restore_Lexer (Loc);
         return No_Node;
      end if;

      Save_Lexer (Loc);
      Restore_Lexer (Properties_Loc);
      Restore_Lexer (Loc);
      return Connection;
   end P_Connection;

end Ocarina.AADL.Parser.Components.Connections;
