------------------------------------------------------------------------------
--                                                                          --
--                           OCARINA COMPONENTS                             --
--                                                                          --
--        O C A R I N A . D I A . P A R S E R . C O R E . N O D E S         --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--               Copyright (C) 2005-2007, GET-Telecom Paris.                --
--                                                                          --
-- Ocarina  is free software;  you  can  redistribute  it and/or  modify    --
-- it under terms of the GNU General Public License as published by the     --
-- Free Software Foundation; either version 2, or (at your option) any      --
-- later version. Ocarina is distributed  in  the  hope  that it will be    --
-- useful, but WITHOUT ANY WARRANTY;  without even the implied warranty of  --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General --
-- Public License for more details. You should have received  a copy of the --
-- GNU General Public License distributed with Ocarina; see file COPYING.   --
-- If not, write to the Free Software Foundation, 51 Franklin Street, Fifth --
-- Floor, Boston, MA 02111-1301, USA.                                       --
--                                                                          --
-- As a special exception,  if other files  instantiate  generics from this --
-- unit, or you link  this unit with other files  to produce an executable, --
-- this  unit  does not  by itself cause  the resulting  executable to be   --
-- covered  by the  GNU  General  Public  License. This exception does not  --
-- however invalidate  any other reasons why the executable file might be   --
-- covered by the GNU Public License.                                       --
--                                                                          --
--                 Ocarina is maintained by the Ocarina team                --
--                       (ocarina-users@listes.enst.fr)                     --
--                                                                          --
------------------------------------------------------------------------------

--  This package contains the definition of the different nodes that may
--  appear in a Dia implementation

with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;
with Ada.Text_IO; use Ada.Text_IO;

with DOM.Core.Nodes;

with Ocarina.Dia.Parser.Core.Ports;
with Ocarina.Dia.Parser.Core.Connections;
with Ocarina.Dia.Parser.Core.Attributes;
with Ocarina.Dia.Parser.Translate.Misc;

package body Ocarina.Dia.Parser.Core.Nodes is

   --  List : Item
   function Item (List : Node_List; Index : Integer) return Node is
   begin
      if Index <= List.Last then
         return List.Items (Index);
      else
         return null;
      end if;
   end Item;

   --  List : Length
   function Length (List : Node_List) return Integer is
   begin
      return List.Last + 1;
   end Length;

   --  Prints a node type
   function Print_Node_Type (Node_Type_Arg : Node_Type) return String is
   begin
      case Node_Type_Arg is
         when Process =>
            return "Process";
         when Thread =>
            return "Thread";
         when Data =>
            return "Data";
         when Processor =>
            return "Processor";
         when Memory =>
            return "Memory";
         when Bus =>
            return "Bus";
         when System =>
            return "System";
         when Subprogram =>
            return "Subprogram";
         when Thread_Group =>
            return "Thread_Group";
         when Device =>
            return "Device";
         when Link =>
            return "Link";
         when Void =>
            return "Void";
      end case;
   end Print_Node_Type;

   --  Prints a node to the screen (for debugging purposes)
   procedure Print_Node (N : Node) is
      use Ocarina.Dia.Parser.Core.Ports;
   begin
      if N = null then
         Put_Line ("<EMPTY NODE>");
      else
         Put_Line ("--- Node ---");
         Put ("> ID : ");
         Put_Line (To_String (N.ID));
         Put ("> Version : ");
         Put_Line (Integer'Image (N.Version));
         Put_Line ("> Type : " & Print_Node_Type (N.Node_Type_Arg));

         case N.Node_Type_Arg is
            when AADLObject_Type =>
               Put_Line ("> Name : " & To_String (N.Name));
               Put_Line ("> Declaration : " & To_String (N.Declaration));
               Put_Line ("> Ports : ");
               Print_Ports (N.Ports);
               Put_Line ("> Parent Node : " & To_String (N.Parent));

            when Link =>
               Put_Line ("> Start Point : ("
                         & To_String (N.Conn.Start_Point.Object_ID)
                         & ", "
                         & Integer'Image (N.Conn.Start_Point.Port)
                         & ")");
               Put_Line ("> End Point : ("
                         & To_String (N.Conn.End_Point.Object_ID)
                         & ", "
                         & Integer'Image (N.Conn.End_Point.Port)
                         & ")");
               Put_Line ("> Style : " & Integer'Image (N.Style));
               Put_Line ("> Head Styles : (" &
                         Integer'Image (N.Start_Head_Style) &
                         " ; " &
                         Integer'Image (N.End_Head_Style) &
                         ")");
            when others =>
               null;
         end case;

         Put_Line ("------------");
      end if;
   end Print_Node;

   --  returns the Node_Type associated with S; raises Not_Matching if none
   function Match_Type (S : String) return Node_Type is
   begin
      if S = "AADL - Process" then
         return Process;
      elsif S =  "AADL - Thread" then
         return Thread;
      elsif S =  "AADL - Data" then
         return Data;
      elsif S =  "AADL - Processor" then
         return Processor;
      elsif S =  "AADL - Memory" then
         return Memory;
      elsif S =  "AADL - Bus" then
         return Bus;
      elsif S =  "AADL - System" then
         return System;
      elsif S =  "AADL - Subprogram" then
         return Subprogram;
      elsif S =  "AADL - Thread Group" then
         return Thread_Group;
      elsif S = "AADL - Device" then
         return Device;
      elsif S =  "Standard - Line"
         or else S = "Standard - Arc"
         or else S = "Standard - ZigZagLine"
         or else S = "Standard - BezierLine" then
         return Link;
      else
         Put_Line (S);
         raise Not_Matching;
      end if;
   end Match_Type;

   --  Reads from the XML/DIA document and returns the next AADL object
   function Get_Node (XMLN : DOM.Core.Node; Id : Natural) return Node is
      use DOM.Core;
      use Ocarina.Dia.Parser.Core.Connections;
      use Ocarina.Dia.Parser.Core.Attributes;
      use Ocarina.Dia.Parser.Core.Ports;

      pragma Unreferenced (Id);

      N : Node;
   begin
      if XMLN.Node_Type = Element_Node
        and then DOM.Core.Nodes.Node_Name (XMLN) = "dia:object"
      then
         declare
            AS : DOM.Core.Named_Node_Map := DOM.Core.Nodes.Attributes (XMLN);
            NT : Node_Type;
            Children : constant DOM.Core.Node_List :=
              DOM.Core.Nodes.Child_Nodes (XMLN);
            E : DOM.Core.Node;
         begin

            NT := Match_Type (Get_Required_Attribute (AS, "type"));
            N := new Node_Record (NT);
            --  id begins with an 'O' and not a '0' => we keep it as a string
            N.ID := To_Unbounded_String (Get_Required_Attribute (AS, "id"));
            N.Version :=
              Integer'Value (Get_Required_Attribute (AS, "version"));

            --  Retrieve the list of attributes
            case NT is
               when AADLObject_Type =>
                  --  Name
                  E := Get_Attribute_Node (Children, "name");
                  N.Name := Get_Name_Attribute (E);
                  --  Declaration
                  E := Get_Attribute_Node (Children, "declaration");
                  N.Declaration := Get_String_Attribute (E);
                  --  Ports
                  N.Ports := Get_Ports (Children);
                  --  Connection Points

                  --  Parent Node
                  declare
                     P : Natural := 0;
                  begin
                     Get_Required_Child (Children, "dia:childnode", P, E);
                     AS := DOM.Core.Nodes.Attributes (E);
                     N.Parent := To_Unbounded_String
                       (Get_Required_Attribute (AS, "parent"));
                  exception
                     when Missing_Child =>
                        null;
                  end;

               when Link =>
                  declare
                     E : DOM.Core.Node;
                     P : Natural := 0;
                  begin
                     --  Connections
                     Get_Required_Child (Children, "dia:connections", P, E);
                     N.Conn := Get_Connection (E);
                     begin
                        --  Style
                        E := Get_Attribute_Node (Children, "line_style");
                        N.Style := Integer'Value
                          (To_String (Get_Val_Attribute (E)));
                     exception
                        when Missing_Dia_Attribute =>
                           N.Style := 0;
                     end;
                     --  Heads' Styles
                     begin
                        E := Get_Attribute_Node (Children, "start_arrow");
                        N.Start_Head_Style := Integer'Value
                          (To_String (Get_Val_Attribute (E)));
                     exception
                        when Missing_Dia_Attribute =>
                           N.Start_Head_Style := 0;
                     end;
                     begin
                        E := Get_Attribute_Node (Children, "end_arrow");
                        N.End_Head_Style := Integer'Value
                          (To_String (Get_Val_Attribute (E)));
                     exception
                        when Missing_Dia_Attribute =>
                           N.End_Head_Style := 0;
                     end;

                  end;
               when others =>
                  null;
            end case;

            return N;
         end;
      else
         return null;
      end if;
   exception
      when Missing_Attribute =>
         return null;
   end Get_Node;

   --  get the name of the node N
   function Get_Node_Name (N : Node) return String is
      use Ocarina.Dia.Parser.Translate.Misc;
   begin
      return Node_Name (To_String (N.Name));
   end Get_Node_Name;

   function Count_Ports (N : Node) return Natural
        is
      Ports : Port_List := N.Ports;
      Compteur : Natural := 0;
   begin
      while Ports /= null loop
         case Ports.Hd.Port_Type_Arg is
            when Real_Port =>
               Compteur := Compteur + 1;
            when others =>
               null;
         end case;
         Ports := Ports.Tl;
      end loop;
      return Compteur;
   end Count_Ports;

   --  returns the number of conections of a node
   function Count_Connections (N : Node) return Natural
   is
      Ports : Port_List := N.Ports;
      Compteur : Natural := 0;
   begin
      while Ports /= null loop
         case Ports.Hd.Port_Type_Arg is
            when Point =>
               Compteur := Compteur + 1;
            when others =>
               null;
         end case;
         Ports := Ports.Tl;
      end loop;
      return Compteur;
   end Count_Connections;

   --  Returns the Id of the node
   function Get_Node_Id (N : Node) return Natural is
   begin
      return Natural'Value (To_String (N.ID));
   end Get_Node_Id;

   --  Returns the Node corresponding to the Id
   function Get_Node_By_Id (Id : Unbounded_String; List : Node_List)
      return Node
   is
      N : Node := null;
   begin
      for I in 0 .. (Length (List) - 1) loop
         N := Item (List, I);
         if N.ID = Id then
            return N;
         end if;
      end loop;
      return null;
   end Get_Node_By_Id;

   ------------------------------
   -- Has_Implementation_Arrow --
   ------------------------------

   function Has_Implementation_Arrow (N : Node; List : Node_List)
      return Node
   is
      O : Node;
   begin
      for I in 0 .. (Length (List) - 1) loop
         O := Item (List, I);
         if O.Node_Type_Arg = Link
            and then O.Style = 4
            and then O.Start_Head_Style = 0
            and then O.End_Head_Style = 2
            and then O.Conn.Start_Point.Object_ID = N.ID
         then
            return Get_Node_By_Id (O.Conn.End_Point.Object_ID, List);
         end if;
      end loop;
      return null;
   end Has_Implementation_Arrow;

end Ocarina.Dia.Parser.Core.Nodes;
