#! /bin/sh

# Script to exploit an AADL model. This script takes a scenario file
# which contains the list of the AADL files to be handled.

# Syntax of a scenario file :
# - The first line of the scenario file contains the system 
#   implementation name.
# - The second line of the scenario file contains the list of the 
#   application nodes to be in the order the have to be run (not used 
#   in this script).
# - The scenario file has to be in the same directory of the AADL models 
#   and the Ada95 source files.

###############################################################################
# Usage

usage () {
   echo "Usage: $0 -s <scenario_file> [-n]";
   echo " scenario_file is the file to be processed"
   echo " -n: to generate Petri Net"
   exit 1;
}

###############################################################################
# Parse command line

parse_cmdline () {
    if [ $# = 0 ]; then
	usage
    fi;

    scenario_file=

    while getopts s:pvn opt
      do
      case $opt in
	  n)
	      GENERATE_PETRI_NET=true
	      ;;
	  s)
	      scenario_file=$OPTARG
	      ;;
	  *)
	      usage
	      ;;
      esac
    done
    shift `expr $OPTIND - 1`

    if [ $scenario_file = "" ]; then
	usage
    fi
}

###############################################################################
# Parse scenario file

parse_scenario () {
    if [ ! -f $scenario_file ]; then
	echo "file $scenario_file not found or not valid"
	exit 1
    fi

    result_dir="`dirname $scenario_file`"
    result_dir=${result_dir}/"`basename $scenario_file .aadl`"

    # Get absolute paths

    work_dir=`pwd`

    source_dir=`dirname $scenario_file`
    cd $source_dir
    source_dir=`pwd`

    cd $work_dir

    # Remove a former generated code
    rm -rf $result_dir

    # Get the files to be handled

    aadl_files="`ocarina_sh -p -s $scenario_file | grep 'Args: -paadl' | awk 'BEGIN {list=""} {for ( i=3 ; i <=NF ; i++ ) {list= list " " $i}}END{print coucou list}'`"
}

###############################################################################
# Generate Petri nets from the scenario

generate_petri_net () {

    echo
    echo "*** Generating Petriscript from the AADL description ***"
    echo

    echo "Parsing files : $aadl_files"
 
    ocarina -c $aadl_files
    a=$?
    if [ $a != 0 ]; then
	echo "error"
	exit $a
    fi

    # Add files for behavioural specifications

    cp "`ocarina -s`/user_defined.psc" "$result_dir"

    # Add behavioural specifications

    cd $source_dir
    for behave_file in `ls *.psc`; do
	tmp=`echo $behave_file | cut -f1 -d'.'` &&
	echo $work_dir/$result_dir
	cp $behave_file $work_dir/$result_dir/$tmp;
    done;
    
    cd $work_dir/$result_dir;
    
    pn_include partition.psc &&
    
    cd $work_dir;
    
}

###############################################################################
# For each call (#include primitive) to a user defined behavioural spec, 
# will build a new instance of behavioural spec, and merge it with main 
# generated script.

pn_include () {

    if [ $# -ne 1 ]; then
	echo "usage : $0 main_file"
	exit 1
    fi

    SRC_INC=$1
    TMP_INC=$SRC_INC.tmp

    touch $TMP_INC
    
    echo "-------------------------------------------------------" >> $TMP_INC
    echo "--                include part                       --" >> $TMP_INC
    echo "--  automatically generated by the 'include' script  --" >> $TMP_INC
    echo "-------------------------------------------------------" >> $TMP_INC
    
    while read cmd arg1 arg2 next;
      do
      if [ "$cmd" = "#include" ];
	  then
	  if [ -e $arg2 ];
	      then
	      pn_namespace $arg1 $arg2 &&
	      cat $arg1'_'$arg2 >> $TMP_INC
	  fi;
      fi;
    done < $SRC_INC &&
    
    echo "-------------------------------------------------------" >> $TMP_INC
    echo "--                generated part                     --" >> $TMP_INC
    echo "-------------------------------------------------------" >> $TMP_INC
    
    echo "" >> $TMP_INC
    
    cat $SRC_INC >> $TMP_INC

    echo "-------------------------------------------------------" >> $TMP_INC
    echo "--                merging part                       --" >> $TMP_INC
    echo "--  automatically generated by the 'include' script  --" >> $TMP_INC
    echo "-------------------------------------------------------" >> $TMP_INC
    
    echo "" >> $TMP_INC
    
    while read cmd arg1 arg2 next;
      do
      if [ "$cmd" = "#include" ];
	  then
	  if [ -e $arg2 ];
	      then
	      rm -f $TMP_INC.tmp &&
	      pn_merge $arg1 $SRC_INC $arg2 $TMP_INC.tmp &&
	      cat $TMP_INC.tmp >> $TMP_INC
	  fi;
      fi;
    done < $SRC_INC &&
    
    # comment all include calls
    cat $TMP_INC | sed s/"#include"/"-- include"/g > $TMP_INC.tmp
    cat $TMP_INC.tmp > $TMP_INC

    rm -f $TMP_INC.tmp;

    mv $TMP_INC generated_final.psc

}

###############################################################################
# For a given call to a user defined behavioural spec, merge all external 
# nodes with corresponding nodes in the generated file.

pn_merge () {

    if [ $# -ne 4 ];
	then
	echo "usage : $0 namespace_str generated_file user_defined_file dest_file"
	exit 1
    fi

    NS_MERGE=$1'_'
    SRC_GEN=$2
    SRC_USR=$3
    TMP_MERGE=$4
    
    # for each node refered by the User defined file, we 
    # put them into the corresponding list

    for var in `grep "append place *" $SRC_USR | cut -f2 -d'"' | cut -f1 -d' ' | sed 's/\("$NS_MERGE"\)\([a-z][a-z0-9_]*\)/\2/'`;
      do     
      echo "merge place "'"'"$var"'"'" and place " '"'"$NS_MERGE$var"'";' >> $TMP_MERGE
    done &&
    
    echo "" >> $TMP_MERGE
    
    for var in `grep "append transition *" $SRC_USR | cut -f2 -d'"' | cut -f1 -d' ' | sed 's/\("$NS_MERGE"\)\([a-z][a-z0-9_]*\)/\2/'`;
      do
      echo "merge transition "'"'"$var"'"'" and transition " '"'"$NS_MERGE$var"'";' >> $TMP_MERGE
    done
    
}

###############################################################################
# Build a petriscript file for a given call of a user defined behavioural 
# spec, prefixing all 'external' (ie. references to existing nodes) by 
# the second argument.

pn_namespace () {

    if [ $# -ne 2 ];
	then
	echo "usage : $0 namespace_str filename"
	exit 1
    fi
    
    NAMESPACE=$1'_'
    SRC_NS=$2
    
    DESTFILE=$NAMESPACE$SRC_NS
    
    echo "--  Namespace $1::$SRC_NS" > $DESTFILE &&
    echo "--  Generated by the 'namespace' script" >> $DESTFILE &&
    
    echo "" >> $DESTFILE &&
    
    TMP_NS=$DESTFILE.tmp
    cat $SRC_NS > $TMP_NS
    
    # replace places
    for var in `grep "create place *" $TMP_NS | cut -f2 -d'"'`;
      do
      cat $TMP_NS | sed s/"$var"/"$NAMESPACE$var"/g > $TMP_NS.tmp2
      cat $TMP_NS.tmp2 > $TMP_NS
    done &&

    # replace transitions
    for var in `grep "create transition *" $TMP_NS | cut -f2 -d'"'`;
      do
      cat $TMP_NS | sed s/"$var"/"$NAMESPACE$var"/g > $TMP_NS.tmp2
      cat $TMP_NS.tmp2 > $TMP_NS
    done &&
    
    # replace list declaration
    cat $TMP_NS | sed s/"places_to_merge"/"${NAMESPACE}places_to_merge"/g > $TMP_NS.tmp2
    cat $TMP_NS.tmp2 > $TMP_NS &&
    cat $TMP_NS | sed s/"transitions_to_merge"/"${NAMESPACE}transitions_to_merge"/g > $TMP_NS.tmp2 &&
    cat $TMP_NS.tmp2 > $TMP_NS
    
    cat $TMP_NS >> $DESTFILE &&
    
    rm -f $TMP_NS.tmp2
    rm -f $TMP_NS
    
    echo "" >> $DESTFILE
    
}

###############################################################################
# The main processing begins here

parse_cmdline $*
parse_scenario

if [ x$GENERATE_PETRI_NET = x"true" ]; then
    generate_petri_net
fi


